package metrics.hyperonymy.potential;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import AST.ASTNode;
import AST.ArrayDecl;
import AST.Block;
import AST.BodyDecl;
import AST.ClassDecl;
import AST.ConstructorDecl;
import AST.FieldDeclaration;
import AST.GenericTypeDecl;
import AST.InstanceInitializer;
import AST.InterfaceDecl;
import AST.MethodDecl;
import AST.NullType;
import AST.ParTypeDecl;
import AST.SimpleSet;
import AST.StaticInitializer;
import AST.TypeDecl;
import metrics.Metric;

// TODO: Mention why "this" is not found
// TODO: Fields of inner classes are not resolved correctly
public class Accessibility implements Metric<BodyDecl> {

	public String getName() {
		return "hyp-pot-acc";
	}

	public void setupHeader(PrintWriter writer) {
		writer.println("project;location;kind of top-level block;varname;kind of variable;static;inherited/fromEnclosing;synthetic;kind of type;type;unique types;no of unique types");
	}

	public void calculate(BodyDecl node) {
		if(node instanceof ConstructorDecl) {
			calculate(node, "constructor");
		} else if(node instanceof InstanceInitializer) {
			calculate(node, "instance initializer");
		} else if(node instanceof MethodDecl) {
			if(!((MethodDecl)node).isAbstract()) {
				calculate(node, "method");
			}
		} else if(node instanceof StaticInitializer) {
			calculate(node, "static initializer");
		} // else: FieldDecl, FieldDeclaration, MemberTypeDecl do not have statements
		  // or should not be handled at this level because they will handled when
		  // collectMetrics() recurses down to the members directly
	}
	
	private void calculate(BodyDecl location, String kindOfTopLevelBlock) {
		// find variables
		SimpleSet variables = location.lookupAllVariables();
		
		// create maps of types and variables
		Map<TypeDecl,Set<AST.Variable>> typeMap = new HashMap<TypeDecl,Set<AST.Variable>>();
		Map<AST.Variable,Set<TypeDecl>> varMap = new HashMap<AST.Variable,Set<TypeDecl>>();
		Iterator<?> iter = variables.iterator();
		while(iter.hasNext()) {
			AST.Variable var = (AST.Variable)iter.next();
			addTypeToMaps(var.type(), var, typeMap, varMap, true);
		}
		
		// log variable accessability
		iter = variables.iterator();
		while(iter.hasNext()) {
			AST.Variable var = (AST.Variable)iter.next();
			logAccessibility(location, var, getUniqueTypes(var, varMap, typeMap),
					location.hostType(), kindOfTopLevelBlock);
		}
	}
	
	private void addTypeToMaps(TypeDecl type, AST.Variable var, Map<TypeDecl,Set<AST.Variable>> typeMap,
			Map<AST.Variable,Set<TypeDecl>> varMap, boolean addInterfaces) {
		// Ignore type parameters since they cannot be expressed as part of identifiers
		if(type.isParameterizedType()) {
			type = ((ParTypeDecl)type).genericDecl();
		}
		
		// add the type itself
		addToTypeMap(typeMap, type, var);
		addToVarMap(varMap, type, var);
		
		// recurse for super-classes
		if(type instanceof ClassDecl) {
			ClassDecl superClass = ((ClassDecl)type).superclass();
			if(superClass != null) {
				addTypeToMaps(superClass, var, typeMap, varMap, false);
			}
		}
		
		// maybe recurse for interfaces
		if(addInterfaces) {
			for(Object interf : type.implementedInterfaces()) {
				addTypeToMaps((InterfaceDecl)interf, var, typeMap, varMap, false);
			}
		}
	}
	
	private java.util.List<String> getUniqueTypes(AST.Variable var, Map<AST.Variable,Set<TypeDecl>> varMap,
			Map<TypeDecl,Set<AST.Variable>> typeMap) {
		java.util.List<String> uniqueTypes = new ArrayList<String>();
		
		for(TypeDecl type : varMap.get(var)) {
			if(!type.fullName().equals("java.lang.Object")
					&& typeMap.get(type).size() == 1) {
				uniqueTypes.add(type.fullName());
			}
		}
		Collections.sort(uniqueTypes);
		return uniqueTypes;
	}
	
	private void addToTypeMap(Map<TypeDecl,Set<AST.Variable>> typeMap, TypeDecl type, AST.Variable var) {
		if(typeMap.containsKey(type)) {
			typeMap.get(type).add(var);
		} else {
			Set<AST.Variable> set = new HashSet<AST.Variable>();
			set.add(var);
			typeMap.put(type, set);
		}
	}
	
	private void addToVarMap(Map<AST.Variable,Set<TypeDecl>> varMap, TypeDecl type, AST.Variable var) {
		if(varMap.containsKey(var)) {
			varMap.get(var).add(type);
		} else {
			Set<TypeDecl> set = new HashSet<TypeDecl>();
			set.add(type);
			varMap.put(var, set);
		}
	}
	
	private void logAccessibility(BodyDecl location, AST.Variable node,
			java.util.List<String> uniqueTypes, TypeDecl hostType, String kindOfDeclaration) {
		String kindOfVariable = Variable.getKindOfVariable(node);
		String kindOfType = Variable.getKindOfType(node.type());
		if(!(node instanceof FieldDeclaration) || ((FieldDeclaration)node).accessibleFrom(hostType)) {
			location.logMetric(this, kindOfDeclaration, node.name(), kindOfVariable, node.isStatic(), 
					isInheritedOrFromEnclosing(hostType, node), node.isSynthetic(), kindOfType,
					node.type().fullName(), uniqueTypes, uniqueTypes.size());
		}
	}
	
	private boolean isInheritedOrFromEnclosing(TypeDecl hostType, AST.Variable var) {
		if(var instanceof FieldDeclaration) {
			return hostType != var.hostType();
		} else {
			return false;
		}
	}

}
